@@ -1099,6 +1099,27 @@ gi_marshalling_tests_utf8_full_inout (gchar **utf8)
 }
 
 
+/**
+ * gi_marshalling_tests_init_function:
+ * @n_args: (inout) (allow-none): number of args
+ * @argv: (inout) (array length=n_args) (allow-none): args
+ *
+ * This is like gtk_init().
+ */
+gboolean
+gi_marshalling_tests_init_function (gint *n_args, char ***argv)
+{
+    if (n_args == NULL)
+	return TRUE;
+
+    if (*n_args == 0)
+	return TRUE;
+    (*n_args)--;
+    g_assert (argv != NULL);
+    (*argv)[*n_args] = NULL;
+    return TRUE;
+}
+
 /**
  * gi_marshalling_tests_array_fixed_int_return:
  * Returns: (array fixed-size=4):
@@ -1315,6 +1336,37 @@ gi_marshalling_tests_array_struct_in (GIMarshallingTestsBoxedStruct **structs, g
     g_assert(structs[2]->long_ == 3);
 }
 
+/**
+ * gi_marshalling_tests_array_simple_struct_in:
+ * @structs: (array length=length):
+ */
+void
+gi_marshalling_tests_array_simple_struct_in (GIMarshallingTestsSimpleStruct *structs, gint length)
+{
+    g_assert(length == 3);
+    g_assert(structs[0].long_ == 1);
+    g_assert(structs[1].long_ == 2);
+    g_assert(structs[2].long_ == 3);
+}
+
+/**
+ * gi_marshalling_tests_multi_array_key_value_in:
+ * @keys: (array length=length):
+ * @values: (array length=length):
+ */
+void
+gi_marshalling_tests_multi_array_key_value_in (gint length, const gchar **keys, const GValue *values)
+{
+    g_assert(length == 3);
+    g_assert(g_strcmp0("one", keys[0]) == 0);
+    g_assert(g_value_get_int(&values[0]) == 1);
+    g_assert(g_strcmp0("two", keys[1]) == 0);
+    g_assert(g_value_get_int(&values[1]) == 2);
+    g_assert(g_strcmp0("three", keys[2]) == 0);
+    g_assert(g_value_get_int(&values[2]) == 3);
+
+}
+
 /**
  * gi_marshalling_tests_array_struct_take_in:
  * @structs: (array length=length) (transfer full):
@@ -1345,29 +1397,6 @@ gi_marshalling_tests_array_enum_in (GIMarshallingTestsEnum *_enum, gint length)
     g_assert (_enum[2] == GI_MARSHALLING_TESTS_ENUM_VALUE3);
 }
 
-/**
- * gi_marshalling_tests_array_nested_in:
- * @list: (array length=length) (element-type GSList<gint>):
- * @length:
- */
-void
-gi_marshalling_tests_array_nested_in (GSList **list, gint length)
-{
-    g_assert (length == 3);
-
-    g_assert (g_slist_length (list[0]) == 1);
-    g_assert (list[0]->data == GINT_TO_POINTER (11));
-
-    g_assert (g_slist_length (list[1]) == 3);
-    g_assert (list[1]->data == GINT_TO_POINTER (21));
-    g_assert (list[1]->next->data == GINT_TO_POINTER (22));
-    g_assert (list[1]->next->next->data == GINT_TO_POINTER (23));
-
-    g_assert (g_slist_length (list[2]) == 2);
-    g_assert (list[2]->data == GINT_TO_POINTER (31));
-    g_assert (list[2]->next->data == GINT_TO_POINTER (32));
-}
-
 /**
  * gi_marshalling_tests_array_in_guint64_len:
  * @ints: (array length=length) (transfer none):
@@ -1549,18 +1578,77 @@ gi_marshalling_tests_array_zero_terminated_inout (gchar ***utf8s)
 }
 
 /**
- * gi_marshalling_tests_array_gvariant_in:
+ * gi_marshalling_tests_array_gvariant_none_in:
  * @variants: (array zero-terminated=1) (transfer none):
+ * Returns: (array zero-terminated=1) (transfer none):
  */
 GVariant **
-gi_marshalling_tests_array_gvariant_in (GVariant **variants)
+gi_marshalling_tests_array_gvariant_none_in (GVariant **variants)
 {
+    /* Use a static container to detect if someone tries to free it */
+    static GVariant *private_container[3] = { NULL, NULL, NULL };
+    
+    if (private_container[0] == NULL) {
+      private_container[0] = g_variant_new_int32 (27);
+      private_container[1] = g_variant_new_string ("Hello");
+    }
+
     g_assert (variants != NULL);
     g_assert_cmpint (g_variant_get_int32 (variants[0]), ==, 27);
     g_assert_cmpstr (g_variant_get_string (variants[1], NULL), ==, "Hello");
     g_assert (variants[2] == NULL);
 
-    return variants;
+    return private_container;
+}
+
+/**
+ * gi_marshalling_tests_array_gvariant_container_in:
+ * @variants: (array zero-terminated=1) (transfer container):
+ * Returns: (array zero-terminated=1) (transfer container):
+ */
+GVariant **
+gi_marshalling_tests_array_gvariant_container_in (GVariant **variants)
+{
+    GVariant **container;
+
+    g_assert (variants != NULL);
+    g_assert_cmpint (g_variant_get_int32 (variants[0]), ==, 27);
+    g_assert_cmpstr (g_variant_get_string (variants[1], NULL), ==, "Hello");
+    g_assert (variants[2] == NULL);
+    
+    container = g_new0 (GVariant*, 3);
+    container[0] = variants[0];
+    container[1] = variants[1];
+    g_free (variants);
+    
+    return container;
+}
+
+/**
+ * gi_marshalling_tests_array_gvariant_full_in:
+ * @variants: (array zero-terminated=1) (transfer full):
+ * Returns: (array zero-terminated=1) (transfer full):
+ */
+GVariant **
+gi_marshalling_tests_array_gvariant_full_in (GVariant **variants)
+{
+    GVariant **container;
+
+    g_assert (variants != NULL);
+    g_assert_cmpint (g_variant_get_int32 (variants[0]), ==, 27);
+    g_assert_cmpstr (g_variant_get_string (variants[1], NULL), ==, "Hello");
+    g_assert (variants[2] == NULL);
+    
+    /* To catch different behaviors we reconstruct one variant from scratch,
+     * while leaving the other untouched. Both approaches are legal with full
+     * transfer in and out */
+    container = g_new0 (GVariant*, 3);
+    container[0] = g_variant_new_int32 (g_variant_get_int32 (variants[0]));
+    g_variant_unref (variants[0]);
+    container[1] = variants[1];
+    g_free (variants);
+
+    return container;
 }
 
 /**
@@ -1805,6 +1893,215 @@ gi_marshalling_tests_garray_utf8_full_inout (GArray **array_)
     *array_ = result;
 }
 
+/**
+ * gi_marshalling_tests_gptrarray_utf8_none_return:
+ * Returns: (element-type utf8) (transfer none):
+ */
+GPtrArray *
+gi_marshalling_tests_gptrarray_utf8_none_return (void)
+{
+    static GPtrArray *parray = NULL;
+    static gchar *values[] = {"0", "1", "2"};
+    gint i;
+
+    if (parray == NULL) {
+        parray = g_ptr_array_new ();
+        for (i = 0; i < 3; i++)
+            g_ptr_array_add (parray, (gpointer) values[i]);
+    }
+
+    return parray;
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_container_return:
+ * Returns: (element-type utf8) (transfer container):
+ */
+GPtrArray *
+gi_marshalling_tests_gptrarray_utf8_container_return (void)
+{
+    GPtrArray *parray = NULL;
+    static gchar *values[] = {"0", "1", "2", NULL};
+    gint i;
+
+    parray = g_ptr_array_new ();
+    for (i = 0; values[i]; i++)
+        g_ptr_array_add (parray, (gpointer)values[i]);
+
+    return parray;
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_full_return:
+ * Returns: (element-type utf8) (transfer full):
+ */
+GPtrArray *
+gi_marshalling_tests_gptrarray_utf8_full_return (void)
+{
+    GPtrArray *parray = NULL;
+    static gchar *values[] = {"0", "1", "2", NULL};
+    gint i;
+
+    parray = g_ptr_array_new ();
+    for (i = 0; values[i]; i++) {
+        gchar *str = g_strdup (values[i]);
+        g_ptr_array_add (parray, (gpointer)str);
+    }
+
+    return parray;
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_none_in:
+ * @parray_: (element-type utf8) (transfer none):
+ */
+void
+gi_marshalling_tests_gptrarray_utf8_none_in (GPtrArray *parray_)
+{
+    g_assert (parray_->len == 3);
+    g_assert (strcmp (g_ptr_array_index (parray_, 0), "0") == 0);
+    g_assert (strcmp (g_ptr_array_index (parray_, 1), "1") == 0);
+    g_assert (strcmp (g_ptr_array_index (parray_, 2), "2") == 0);
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_none_out:
+ * @parray_: (out) (element-type utf8) (transfer none):
+ */
+void
+gi_marshalling_tests_gptrarray_utf8_none_out (GPtrArray **parray_)
+{
+    static GPtrArray *internal = NULL;
+    static gchar *values[] = {"0", "1", "2", NULL};
+    gint i;
+
+    if (internal == NULL) {
+        internal = g_ptr_array_new ();
+        for (i = 0; values[i]; i++)
+            g_ptr_array_add (internal, (gpointer)values[i]);
+    }
+
+    *parray_ = internal;
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_container_out:
+ * @parray_: (out) (element-type utf8) (transfer container):
+ */
+void
+gi_marshalling_tests_gptrarray_utf8_container_out (GPtrArray **parray_)
+{
+    static gchar *values[] = {"0", "1", "2", NULL};
+    gint i;
+
+    *parray_ = NULL;
+
+    *parray_ = g_ptr_array_new ();
+    for (i = 0; values[i]; i++)
+        g_ptr_array_add (*parray_, (gpointer)values[i]);
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_full_out:
+ * @parray_: (out) (element-type utf8) (transfer full):
+ */
+void
+gi_marshalling_tests_gptrarray_utf8_full_out (GPtrArray **parray_)
+{
+    static gchar *values[] = {"0", "1", "2", NULL};
+    gint i;
+
+    *parray_ = NULL;
+
+    *parray_ = g_ptr_array_new ();
+    for (i = 0; values[i]; i++) {
+        gchar *str = g_strdup (values[i]);
+        g_ptr_array_add (*parray_, (gpointer)str);
+    }
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_none_inout:
+ * @parray_: (inout) (element-type utf8) (transfer none):
+ */
+void
+gi_marshalling_tests_gptrarray_utf8_none_inout (GPtrArray **parray_)
+{
+    static GPtrArray *internal = NULL;
+    static gchar *values[] = {"-2", "-1", "0", "1", NULL};
+    gint i;
+
+    g_assert ((*parray_)->len == 3);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 0), "0") == 0);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 1), "1") == 0);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 2), "2") == 0);
+
+    if (internal == NULL) {
+        internal = g_ptr_array_new ();
+        for (i = 0; values[i]; i++)
+            g_ptr_array_add (internal, (gpointer) values[i]);
+    }
+
+    *parray_ = internal;
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_container_inout:
+ * @parray_: (inout) (element-type utf8) (transfer container):
+ */
+void
+gi_marshalling_tests_gptrarray_utf8_container_inout (GPtrArray **parray_)
+{
+    static gchar *val1 = "-2";
+    static gchar *val2 = "-1";
+    static gchar *val3 = "0";
+    static gchar *val4 = "1";
+    GPtrArray *result;
+
+    g_assert ((*parray_)->len == 3);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 0), "0") == 0);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 1), "1") == 0);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 2), "2") == 0);
+
+    result = g_ptr_array_new ();
+    g_ptr_array_add (result, (gpointer) val1);
+    g_ptr_array_add (result, (gpointer) val2);
+    g_ptr_array_add (result, (gpointer) val3);
+    g_ptr_array_add (result, (gpointer) val4);
+
+    *parray_ = result;
+}
+
+/**
+ * gi_marshalling_tests_gptrarray_utf8_full_inout:
+ * @parray_: (inout) (element-type utf8) (transfer full):
+ */
+void
+gi_marshalling_tests_gptrarray_utf8_full_inout (GPtrArray **parray_)
+{
+    static gchar *val1 = "-1";
+    static gchar *val2 = "-2";
+    gchar *val;
+    GPtrArray *result;
+
+    g_assert ((*parray_)->len == 3);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 0), "0") == 0);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 1), "1") == 0);
+    g_assert (strcmp (g_ptr_array_index (*parray_, 2), "2") == 0);
+
+    result = g_ptr_array_new ();
+    val = g_strdup (val2);
+    g_ptr_array_add(result, (gpointer) val);
+    val = g_strdup (val1);
+    g_ptr_array_add(result, (gpointer) val);
+    val = g_strdup ("0");
+    g_ptr_array_add(result, (gpointer) val);
+    val = g_strdup ("1");
+    g_ptr_array_add(result, (gpointer) val);
+
+    *parray_ = result;
+}
+
 /**
  * gi_marshalling_tests_bytearray_full_return:
  * Returns: (transfer full):
@@ -1813,14 +2110,10 @@ GByteArray *
 gi_marshalling_tests_bytearray_full_return (void)
 {
     GByteArray *array = NULL;
-    static gchar *values[] = {"0", "1", "2", "3", NULL};
-    gint i;
+    guint8 data[] = {'\0', '1', '\xFF', '3'};
 
     array = g_byte_array_new ();
-    for (i = 0; values[i]; i++) {
-        gchar *str = g_strdup (values[i]);
-        g_byte_array_append (array, (const guint8*)str, 1);
-    }
+    g_byte_array_append (array, (const guint8*)data, G_N_ELEMENTS(data));
 
     return array;
 
@@ -1828,16 +2121,16 @@ gi_marshalling_tests_bytearray_full_return (void)
 
 /**
  * gi_marshalling_tests_bytearray_none_in:
- * @array_: (element-type gint) (transfer none):
+ * @array_: (element-type gint8) (transfer none):
  */
 void
 gi_marshalling_tests_bytearray_none_in (GByteArray *array_)
 {
-    g_assert (array_->len == 4);
-    g_assert (g_array_index (array_, char, 0) == '0');
-    g_assert (g_array_index (array_, char, 1) == '1');
-    g_assert (g_array_index (array_, char, 2) == '2');
-    g_assert (g_array_index (array_, char, 3) == '3');
+    g_assert_cmpuint (array_->len, ==, 4);
+    g_assert_cmpuint (g_array_index (array_, unsigned char, 0), ==, 0);
+    g_assert_cmpuint (g_array_index (array_, unsigned char, 1), ==, 49);
+    g_assert_cmpuint (g_array_index (array_, unsigned char, 2), ==, 0xFF);
+    g_assert_cmpuint (g_array_index (array_, unsigned char, 3), ==, 51);
 }
 
 /**
@@ -3440,10 +3733,20 @@ gi_marshalling_tests_sub_object_finalize (GObject *object)
 	G_OBJECT_CLASS(gi_marshalling_tests_sub_object_parent_class)->finalize(object);
 }
 
+static void
+method_deep_hierarchy (GIMarshallingTestsObject *self, gint8 in)
+{
+    GValue val = {0, };
+    g_value_init (&val, G_TYPE_INT);
+    g_value_set_int (&val, in);
+    g_object_set_property (G_OBJECT (self), "int", &val);
+}
+
 static void
 gi_marshalling_tests_sub_object_class_init (GIMarshallingTestsSubObjectClass *klass)
 {
 	G_OBJECT_CLASS(klass)->finalize = gi_marshalling_tests_sub_object_finalize;
+    GI_MARSHALLING_TESTS_OBJECT_CLASS(klass)->method_deep_hierarchy = method_deep_hierarchy;
 }
 
 void
@@ -3458,6 +3761,18 @@ gi_marshalling_tests_sub_object_overwritten_method (GIMarshallingTestsSubObject
     g_assert(GI_MARSHALLING_TESTS_OBJECT(object)->int_ == 0);
 }
 
+G_DEFINE_TYPE (GIMarshallingTestsSubSubObject, gi_marshalling_tests_sub_sub_object, GI_MARSHALLING_TESTS_TYPE_SUB_OBJECT);
+
+static void
+gi_marshalling_tests_sub_sub_object_init (GIMarshallingTestsSubSubObject *object)
+{
+}
+
+static void
+gi_marshalling_tests_sub_sub_object_class_init (GIMarshallingTestsSubSubObjectClass *klass)
+{
+}
+
 /* Interfaces */
 
 static void
